import time,ctypes,json,sys
_ = {'_':time.time(),'version':20210401.3}
import win32api,win32con
from win32process import GetWindowThreadProcessId
from win32gui import \
    GetClassName,GetWindowText,GetWindowRect,\
    ClientToScreen,ScreenToClient,SendMessage,\
    GetDesktopWindow,EnumWindows,IsWindowVisible,\
    WindowFromPoint,GetCursorPos,ChildWindowFromPointEx,\
    GetWindowPlacement,ShowWindow,SetForegroundWindow,\
    PostMessage,EnumChildWindows,IsWindowEnabled
from ctypes import windll,wintypes
#import ctypes
#from ctypes.wintypes import MSG    #引入ctypes的包，实现调用windows的API
#import win32gui,win32con
# import array,struct
import win32clipboard

sys_import = __import__
flag_py2 = sys.version_info.major==2
sys_reload = __builtins__.reload if flag_py2 else sys_import('importlib').reload
refresh = lambda:sys_reload(sys_import(__name__))

def tryx(l,e=print):
    try: return l()
    except Exception as ex: return ex if True==e else e(ex) if e else None

cls_name = lambda h:tryx(lambda:GetClassName(h))
win_text = lambda h:tryx(lambda:GetWindowText(h))
win_rect = lambda h:tryx(lambda:GetWindowRect(h))
#clt_rect = lambda h:tryx(lambda:GetClientRect(h))
client2screen=lambda h,p:tryx(lambda:ClientToScreen(h,p))
screen2client=lambda h,p:tryx(lambda:ScreenToClient(h,p))
htc = lambda h:(h,win_text(h),cls_name(h))
tpid = lambda h:tryx(lambda:tuple(GetWindowThreadProcessId(h)))
def win_text_set(h,t):
    tryx(lambda:SendMessage(h, win32con.WM_SETTEXT, None, str(t)))
    return myself

psub = lambda p1,p2:(p1[0]-p2[0],p1[1]-p2[1])
padd = lambda p1,p2:(p1[0]+p2[0],p1[1]+p2[1])

def sleep(t):
    time.sleep(t)
    return myself

# fix for main-window (who might have a win-theme-border)
def win_rect_fix(h):
    if not h: h = GetDesktopWindow()
    rect = wintypes.RECT()
    DWMWA_EXTENDED_FRAME_BOUNDS = 9
    # DwmGetWindowAttribute(h,&rect)
    windll.dwmapi.DwmGetWindowAttribute(wintypes.HWND(h),
      wintypes.DWORD(DWMWA_EXTENDED_FRAME_BOUNDS),
      ctypes.byref(rect),
      ctypes.sizeof(rect))
    _l,_t,_r,_b = (rect.left,rect.top,rect.right,rect.bottom)
    return (_l,_t,_r,_b) if _l or _t or _r or _b else win_rect(h)

rh_root = lambda h:windll.user32.GetAncestor(h,win32con.GA_ROOT)
rhtc = lambda h:htc(rh_root(h))

# rh_rootowner = lambda h:windll.user32.GetAncestor(h,win32con.GA_ROOTOWNER)
# Out: [(hdl,txt,cls,tid,pid,rh),...]
# def main_windows(reverse=True,visible=True):
#     rt = []
#     EnumWindows(lambda _h,_p:_p.append(
#         htc(_h)+tpid(_h)+(rh_rootowner(_h),)
#         ) if visible==None or visible==IsWindowVisible(_h) else None, rt)
#     if reverse: rt.reverse()
#     return rt

# pos@screen => pos@h
hp2p = lambda h,p:psub(p,win_rect_fix(h))

#### pos@screen => hwnd-of-ctrl
p2htc = lambda p=None:tryx(lambda:htc(WindowFromPoint(p or GetCursorPos())))

#### pos@win => hdl (shallow)
wp2h0 = lambda h,p,f=win32con.CWP_SKIPINVISIBLE:tryx(
    lambda:ChildWindowFromPointEx(h,p,f))

def wp2h1(h,p):
    a = win_find(h,x=p[0],y=p[1])
    if len(a)>0: return a[-1][0]
    return 0

#### win.pos => hdl (deep)
# WARN: wrong when minimized
def wp2htc(ph,p):
    # return htc(wp2h1(ph,p))
    rt = ph
    _pp = p
    _rect_ph = win_rect_fix(ph)
    while rt:
        _p = screen2client(rt,p)
        _pp = padd(_p,_rect_ph)
        _h = wp2h0(rt,_pp)
        #print('dbg.wp2htc',rt,p,_p,'pp',_pp,_rect_ph,_h)
        if not _h or _h==rt: break
        rt = _h
    if not rt:
        rt = wp2h1(ph,p) # backup plan (not yet 100% correct)
        print('dbg.wp2htc planb',htc(rt))
    return htc(rt)+_pp

#flags, state, ptMin, ptMax, rect =
win_state=lambda h:tryx(lambda:GetWindowPlacement(h or GetDesktopWindow()))

def win_change(h,fore=False,mx=None,mn=None,nm=None,rect=False,sleep=0.2):
  if mn==True: ShowWindow(h,win32con.SW_SHOWMINIMIZED)
  if mx==True: ShowWindow(h,win32con.SW_SHOWMAXIMIZED)
  if nm==True: ShowWindow(h,win32con.SW_SHOWNORMAL)
  if fore==True: tryx(lambda:SetForegroundWindow(h))
  time.sleep(sleep)
  return win_rect_fix(h)

def win_close(h):
    tryx(lambda:PostMessage(h, win32con.WM_CLOSE,0,0))
    return myself

def win_kill(h):
    tryx(lambda:PostMessage(h, win32con.WM_DESTROY,0,0))
    return myself

def win_focus(h, sleep=0.02):
    SendMessage(h, win32con.WM_ACTIVATE, win32con.WA_ACTIVE, 0)
    #time.sleep(sleep)
    return myself

def win_key_x(h,x,sleep_dn=0.02,sleep_up=0,mode=0):
    print(x)#tmp quick debug
    if mode==0:
        PostMessage(h, win32con.WM_KEYDOWN, x, 0)
    else:
        ctypes.windll.user32.keybd_event(x, 0, 0, 0)
    if sleep_up>0: time.sleep(sleep_up)
    if mode==0:
        SendMessage(h, win32con.WM_KEYUP, x, 0)
    else:
        ctypes.windll.user32.keybd_event(x, 0, 2, 0)
    if sleep_dn>0: time.sleep(sleep_up)
    return myself

def win_key_c(h,c,sleep=0.02,mode=0):
    win_key_x(h,ord(c),sleep_up=sleep,mode=mode)
    return myself

def win_input_x(h,x,z=0,sleep=0.02):
    SendMessage(h, win32con.WM_CHAR,x,z)
    return myself
    
def win_input_c(h,u,sleep=0.02):
    return win_input_x(h,ord(u),sleep)

def ctrl_key(win,key=86,xkey=17):
    win32api.keybd_event(xkey, 0, 0, 0) #ctrl(17)
    time.sleep(0.5)
    SendMessage(win, win32con.WM_KEYDOWN, key, 0)
    win32api.keybd_event(xkey, 0, win32con.KEYEVENTF_KEYUP, 0)

# NOTES: tried, still popup verify-code in ths...
def ctrl_copy(h):
    PostMessage(h,win32con.WM_COMMAND, 0xE122, 0)

# def win_event(eventType, msg):#PYQT里的函数，固定的
#     mess=  ctypes.wintypes.MSG.from_address(msg.__int__())
#     if mess.message==0x004A:
#         PCOPYDATASTRUCT = ctypes.POINTER(COPYDATASTRUCT)
#         pCDS = ctypes.cast(mess.lParam, PCOPYDATASTRUCT)
#         s=ctypes.string_at(pCDS.contents.lpData)
#         print(s.decode('utf-8'))
#     return False, 0

# TODO https://github.com/NT5/pyhelper/blob/1e229792a9057e2dfad5810094fee755194f4066/_old/beta%20concep/res/osu_np.py
#https://github.com/scsnake/miniPACS/blob/1de912a6109c0e572e9c17305f75f4dad07fd02a/win32func.py
# class COPYDATASTRUCT(ctypes.Structure):
#     _fields_ = [
#         ('dwData', ctypes.wintypes.LPARAM),
#         ('cbData', ctypes.wintypes.DWORD),
#         ('lpData', ctypes.c_wchar_p)
#     ]
# PCOPYDATASTRUCT = ctypes.POINTER(COPYDATASTRUCT)
# def win_send(hwnd, str, dwData=0):
#     cds = COPYDATASTRUCT()
#     cds.dwData = dwData
#     cds.cbData = ctypes.sizeof(ctypes.create_unicode_buffer(str))
#     cds.lpData = ctypes.c_wchar_p(str)

#     return ctypes.windll.user32.SendMessageW(hwnd, win32con.WM_COPYDATA, 0, ctypes.byref(cds))    
#     # int_buffer = array.array("L", [0])
#     # char_buffer = array.array('b', b"abcdefg")
#     # int_buffer_address = int_buffer.buffer_info()[0]
#     # #char_buffer = array.array('B',("12341234").encode('utf-8'))
#     # char_buffer_address, char_buffer_size = char_buffer.buffer_info()
#     # #copy_struct = struct.pack("PLP", 1, char_buffer_size, char_buffer_address)
#     # copy_struct = struct.pack("PLP", int_buffer_address, char_buffer_size, char_buffer_address)
#     # SendMessage(h, win32con.WM_COPYDATA, None, copy_struct)
#     # return copy_struct
 
# click pos@ctrl
def win_click_ctrl(h, l_or_r, p):
    _long =win32api.MAKELONG(p[0],p[1])
    _wm_dn = win32con.WM_LBUTTONDOWN if l_or_r==1 else win32con.WM_RBUTTONDOWN
    _wm_up = win32con.WM_LBUTTONUP if l_or_r==1 else win32con.WM_RBUTTONUP
    _mk = win32con.MK_LBUTTON if l_or_r==1 else win32con.MK_RBUTTON
    #PostMessage(h,win32con.WM_MOVE,0,_long) # important, move little
    PostMessage(h,win32con.WM_MOUSEMOVE,0,_long) # important, move little
    PostMessage(h, _wm_dn, _mk, _long)
    PostMessage(h, _wm_up, _mk, _long) # no SendMessage
    return myself

# click pos@win
def win_click(ph,l_or_r,pos_at_win,debug=False):
    h,t,c,px,py = wp2htc(ph,pos_at_win)
    if debug: print('win_click',h,t,c,ph,pos_at_win)
    pos_at_screen = padd(pos_at_win,win_rect_fix(ph))
    pos_at_ph = hp2p(h,pos_at_screen)
    # print(ph,pos_at_win,'=>',h,pos_at_ph)
    return win_click_ctrl(h,l_or_r,pos_at_ph)

def rgb(pixel): return pixel & 0x0000ff, (pixel & 0x00ff00) >> 8, pixel >> 16

def probe_color(pos, h=0):
    if h==0:
        hdc = windll.user32.GetDC(0)
        pixel = windll.gdi32.GetPixel(hdc, pos[0], pos[1])
    else:
        h,t,c,px,py = wp2htc(h,pos)
        hdc = windll.user32.GetDC(h)
        pixel = windll.gdi32.GetPixel(hdc, px, py)
    return rgb(pixel) + (pixel,)
    
#def get_color(p):
#    hdc = windll.user32.GetDC(0)
#    pixel = windll.gdi32.GetPixel(hdc, p[0], p[1])  # 提取RGB值
#    return rgb(pixel) + (pixel,)

#def get_color_win(ph, pos_at_win):
#    h,t,c,px,py = wp2htc(ph,pos_at_win)
#    hdc = windll.user32.GetDC(h)
#    pixel = windll.gdi32.GetPixel(hdc, px, py)  # 提取RGB值
#    return rgb(pixel) + (pixel,)

#api = ctypes.windll.user32
#api.GetDlgItemTextW(api.GetParent(lv.handle), 1576, buff, 96)
def listview_data(lv, on=True):
    api = ctypes.windll.user32
    buff = ctypes.create_unicode_buffer(96)
    #lv = self._client['SysListView32']
    #time.sleep(0.5)
    if on:
        raw = [x.text() for x in lv.items()]
        return list(zip(*[iter(raw)] * lv.column_count()))
    api.GetDlgItemTextW(api.GetParent(lv.handle), 1576, buff, 96)
    print('buff.value=',buff.value)
    return dict([x.split(':') for x in buff.value.strip().split('  ')])

# #import time
# import struct
# #import win32api
# #import win32gui
# #import ctypes
# #import win32clipboard
# #import win32con
# import commctrl
# user32_GetWindowThreadProcessId = ctypes.windll.user32.GetWindowThreadProcessId
# VirtualAllocEx = ctypes.windll.kernel32.VirtualAllocEx
# VirtualFreeEx = ctypes.windll.kernel32.VirtualFreeEx
# OpenProcess = ctypes.windll.kernel32.OpenProcess
# WriteProcessMemory = ctypes.windll.kernel32.WriteProcessMemory
# ReadProcessMemory = ctypes.windll.kernel32.ReadProcessMemory
# 
# def _readListViewItems(hwnd, column_index=0):
#     # Allocate virtual memory inside target process
# 
#     #pid = ctypes.create_string_buffer(4)
#     pid = ctypes.create_string_buffer(4)
#     p_pid = ctypes.addressof(pid)
#     print('debug pid',pid)
#     user32_GetWindowThreadProcessId(hwnd, p_pid)  # process owning the given hwnd
# 
#     #tid,pid = tpid(hwnd)
# 
#     hProcHnd = OpenProcess(win32con.PROCESS_ALL_ACCESS, False, struct.unpack("i", pid)[0])
#     pLVI = VirtualAllocEx(hProcHnd, 0, 4096, win32con.MEM_RESERVE | win32con.MEM_COMMIT, win32con.PAGE_READWRITE)
#     pBuffer = VirtualAllocEx(hProcHnd, 0, 4096, win32con.MEM_RESERVE | win32con.MEM_COMMIT, win32con.PAGE_READWRITE)
# 
#     # Prepare an LVITEM record and write it to target process memory
#     lvitem_str = struct.pack('iiiiiiiii', *[0, 0, column_index, 0, 0, pBuffer, 4096, 0, 0])
#     lvitem_buffer = ctypes.create_string_buffer(lvitem_str)
#     copied = ctypes.create_string_buffer(4)
#     p_copied = ctypes.addressof(copied)
#     WriteProcessMemory(hProcHnd, pLVI, ctypes.addressof(lvitem_buffer), ctypes.sizeof(lvitem_buffer), p_copied)
# 
#     # iterate items in the SysListView32 control
#     num_items = win32gui.SendMessage(hwnd, commctrl.LVM_GETITEMCOUNT)
#     item_texts = []
#     for item_index in range(num_items):
#         win32gui.SendMessage(hwnd, commctrl.LVM_GETITEMTEXT, item_index, pLVI)
#         target_buff = ctypes.create_string_buffer(4096)
#         ReadProcessMemory(hProcHnd, pBuffer, ctypes.addressof(target_buff), 4096, p_copied)
#         item_texts.append(target_buff.value)
# 
#     VirtualFreeEx(hProcHnd, pBuffer, 0, win32con.MEM_RELEASE)
#     VirtualFreeEx(hProcHnd, pLVI, 0, win32con.MEM_RELEASE)
#     win32api.CloseHandle(hProcHnd)
#     return item_texts
# 
# def getListViewInfo(hwnd, cols):
#     """
#     获取ListView的信息
#     :param hwnd: sysListView句柄
#     :param cols: 读取的列数
#     :return: sysListView中的内容
#     """
#     col_info = []
#     for col in range(cols):
#         col_info.append(_readListViewItems(hwnd, col))
#     row_info = []
# 
#     # 按行
#     for row in range(len(col_info[0])):
#         row_info.append([])
#         for col in range(len(col_info)):
#             row_info[row].append(col_info[col][row].decode('GB2312'))
#     return row_info

def win_find(ph=None,pid=None,ha=None,visible=None,enable=None,
             cls=None,clsa=None,txt=None,x=None,y=None):
    if ph: tid,pid = tpid(ph)
    elif ph==0:
        #print('WARN finding root ph=0')
        ''
    elif not pid:
        print('ERR: win_find() needs ph or pid')
        return None
    if ph!=None:
        pha = [ph]
    else:# using pid to filter
        pha = []
        def _filter0(_h,_p):
            #if not IsWindowVisible(_h): return
            _tid,_pid = tpid(_h)
            if pid == _pid:
                # print('tmp',pid,_pid,_h,cls_name(_h),win_text(_h))
                _p.append(_h)
        EnumChildWindows(0,_filter0,pha)
        pha.reverse()
    # print('pha',pha,'pid',pid)
    rt=[]
    for _ph in pha:
        ph_x=0
        ph_y=0
        if x and y:
            if _ph>0: # not desktop
                _rect = win_rect_fix(_ph)
                print('dbg.win_find():_rect,_ph',_rect,_ph)
                ph_x,ph_y,ph_r,ph_b = _rect
        def _filter(_h,p):
          flag_append = True
          _txt=win_text(_h)
          _cls=cls_name(_h)
          _tid,_pid = tpid(_h)
          if ha and _h not in ha: flag_append = False
          _visible = IsWindowVisible(_h)
          if None!=visible and visible!=_visible: flag_append = False
          _enable = IsWindowEnabled(_h)
          if None!=enable and enable!=_enable: flag_append = False
          if pid and pid!=_pid: flag_append = False
          if flag_append:
              if clsa and (_cls not in clsa):
                  flag_append = False
          if cls and cls!=_cls: flag_append = False
          if txt!=None and txt!=_txt: flag_append = False
          # todo: what if using win_rect_fix??
          # todo: any improvement here...
          rect = win_rect(_h)
          ww = rect[2]-rect[0]
          hh = rect[3]-rect[1]
          if flag_append and x and y:
              if _h==0: # desktop skip
                  flag_append = True
              else:
                  xx = x + ph_x
                  yy = y + ph_y
                  if not (xx>=rect[0] and xx<=rect[2] and yy>=rect[1] and yy<=rect[3]):
                      # print('X',_h,rect,(xx,x,ph_x),(yy,y,ph_y),_txt,_cls)
                      flag_append = False
                  # else: print('=>',_h,rect,xx,yy,_txt,_cls)
          if flag_append:
              p.append((_h,_txt,_cls,_pid,_visible,_enable,_ph,rect,(ww,hh)))
          #elif _visible: print('?',_h,_txt,_cls,clsa)
        EnumChildWindows(_ph,_filter,rt)
    return rt

def mon():
    prev_pos = None
    while True: # quick spy ;)
        time.sleep(1)
        screen_pos = win32api.GetCursorPos()

        # skip if no move
        if screen_pos == prev_pos: continue

        #_hdl,_txt,_cls = p2htc()
        _hdl,_txt,_cls = p2htc(screen_pos)
        tid,pid = tpid(_hdl)

        ph,p_txt,p_cls = rhtc(_hdl)
        ptid,ppid = tpid(ph)

        rect = win_rect_fix(ph)
        
        pos_ctrl_at_win = hp2p(ph,screen_pos)

        cch2,cch2_txt,cch2_cls,px,py = wp2htc(ph,pos_ctrl_at_win) 

        print('\nscreen',screen_pos,_hdl,_cls,_txt,pid,probe_color(screen_pos),
                '\n win',rect,ph,p_cls,p_txt,ppid,
                '\n pos@win',pos_ctrl_at_win,cch2,cch2_cls,cch2_txt,
                '\n pos@ctrl',(px,py), probe_color(pos_ctrl_at_win,ph),
            )
        prev_pos = screen_pos

# wins = lambda ph=[0],ha=None,cls=None,txt=None,full=False,visible=True,enable=True:[ v if full else v[0] for _h in ph for v in win_find(ph=_h,ha=ha,cls=cls,txt=txt)]
# wins = lambda ph=[0],ha=None,cls=None,clsa=None,txt=None,full=False,visible=True,enable=False:[ v if full else v[0] for _cls in (clsa if clsa else [cls]) for _h in ph for v in win_find(ph=_h,ha=ha,cls=_cls,txt=txt)]
wins = lambda ph=[0],ha=None,cls=None,clsa=None,txt=None,full=False,visible=None,enable=None:[ v if full else v[0] for _h in ph for v in win_find(ph=_h,ha=ha,clsa=clsa if clsa else [cls] if cls else None,txt=txt)]

json_dumps = lambda o,format=True,indent=None:json.dumps(o,ensure_ascii=False,indent=indent)
list_dumps = lambda l:len([print(v) for v in l])

def input(h, s, sleep=0.02, focus=True, clear=False):
    if focus:
        win_focus(h)
        time.sleep(sleep)
    if clear:
        win_text_set(h, '')
    for c in str(s): win_input_c(h,c,sleep)
    return myself

def paste(h, txt, pos = (0,0), focus=False, restore=False, mode='ctrlv', debug=False):
    if focus: win_focus(h)

    if mode=='input' or tryx(lambda:int(txt),False) is not None: # fix bug about int
      input(h,txt)
    else:
      txt = str(txt)
      win32clipboard.OpenClipboard()
      if restore: _old= win32clipboard.GetClipboardData()
      win32clipboard.EmptyClipboard()
      win32clipboard.SetClipboardData(win32con.CF_UNICODETEXT, txt)
      win32clipboard.CloseClipboard()
      if mode=='paste':# TODO
          #tryx(lambda:SendMessage(h,win32con.WM_CHAR,22,2080193))
          _long =win32api.MAKELONG(pos[0],pos[1])
          if debug: print('paste',_long)
          tryx(lambda:SendMessage(h,win32con.WM_PASTE,0,_long))
      else:
          ctrl_key(h,key=86) #ctrl-v = 86
      if restore:
          win32clipboard.OpenClipboard()
          win32clipboard.EmptyClipboard()
          win32clipboard.SetClipboardData(win32con.CF_UNICODETEXT, _old)
          win32clipboard.CloseClipboard() 
    return myself
    
########################################## chain pattern
class bot:
    def __init__(self,clsa=None,h=None,ha=None):
        self.clsa = clsa
        self.h = h if h else 0
        self.ha = ha

    #usage:  h,*h_rest = fuyi._ha()
    #def _ha(self): return self.ha if self.ha else wins(clsa=self.clsa) if self.clsa else [self.h]
    def _ha(self):
        if self.ha:
            return self.ha
        elif self.clsa:
            return wins(clsa=self.clsa)
        else:
            return [self.h]

    # for quick dump print(bot)
    def __repr__(self):
        _ha = self._ha()
        rt = []
        for h in _ha:
          _txt=win_text(h)
          _cls=cls_name(h) if h>0 else None
          rt.append((h,_cls,_txt))
        return json.dumps(rt,ensure_ascii=False)

    def __getitem__(self,k):# []
        _ha = self._ha()
        rt = []
        c = 0
        for h in _ha:
            #if c==k: return [h,win_text(h),cls_name(h)]
            if c==k: return htc(h)
            c +=1

    def __len__(self): return len(self._ha())
    def len(self): return len(self._ha())

    def get00(self): return tryx(lambda:self[0][0])

    def fore(self,mx=False,sleep=0.2): # todo 
        for _h in self._ha(): win_change(_h,fore=True,mx=mx)
        time.sleep(sleep)
        return self

    def wins(self,ph=None,idx=0,clsa=None,cls=None,txt=None):
        _wins = self._ha()
        print('wins()',_wins)
        if clsa or cls or txt:
            _winsx = wins(ph=ph or _wins,clsa=clsa,cls=cls,txt=txt)
            _wins = _winsx
            print('wins()',_wins)
            if idx>0:
                c = 0
                for h in _wins:
                    c += 1
                    if c==idx: return [h]
        return _wins

    def click(self, l_or_r,p,idx=0,clsa=None,cls=None,txt=None,sleep=0.2):
        _wins = self._ha()

        print('click() _wins',_wins)
        if clsa or cls or txt:
            _winsx = wins(ph=_wins,clsa=clsa,cls=cls,txt=txt)
            #if len(_winsx)>0:
            _wins = _winsx
            print('click() _wins',_wins)

        flag_do_sleep = False
        c = 0
        for h in _wins:
            c += 1
            if idx>0:
                if c==idx:
                    win_click(h,l_or_r,p)
                    break
            else:
                win_click(h,l_or_r,p)
                flag_do_sleep = True

        if flag_do_sleep and sleep>0:
            time.sleep(sleep)

        return self

    def sleep(self, t): time.sleep(t); return self

    def typewrite(self, s):
        import pyautogui
        pyautogui.typewrite(str(s))
        return self

    def input(self,s,cls=None,txt=None,idx=0,sleep=0.2,clear=True):
        _wins = self._ha()
        print('input() _wins',_wins)
        if cls or txt:
            _winsx = wins(ph=_wins,cls=cls,txt=txt)
            #if len(_winsx)>0: _wins = _winsx
            _wins = _winsx
            print('input() _wins',_wins)
        c = 0
        for h in _wins:
            c += 1
            if idx>0:
                if idx==c:
                    input(h,s,sleep,clear=clear)
                    break
            else:
                input(h,s,sleep,clear=clear)
        return self

    def keys(self,s,cls=None,txt=None,idx=0,sleep=0.01,mode=0):

        _wins = self._ha()
        print('keys() _wins',_wins)
        if cls or txt:
            _winsx = wins(ph=_wins,cls=cls,txt=txt)
            #if len(_winsx)>0: _wins = _winsx
            _wins = _winsx
            print('keys() _wins',_wins)
        c = 0
        for h in _wins:
            c += 1
            if idx>0:
                if idx == c:
                    win_focus(h)
                    for cc in str(s): win_key_c(h,cc,sleep=sleep,mode=mode)
                    break
            else:
                #key(h,s,sleep)
                win_focus(h)
                for cc in str(s): win_key_c(h,cc,sleep=sleep,mode=mode)
        return self

    def txt(self,txt,cls,idx):
        _wins = self._ha()
        if cls:
            _winsx = wins(ph=_wins,cls=cls)
            if len(_winsx)>0: _wins = _winsx
        print('txt() _wins',_wins)
        c = 0
        for h in _wins:
            c += 1
            if idx>0 and idx==c:
                win_text_set(h,txt)
                break
        return self

def test(req=None, ph=0, debug=False,cls=None,txt=None):

  def _chk(h):
    try:
      _cls=cls_name(h)
      if _cls=='TMainForm':win_text_set(h,'fuyi')
      if _cls=='#32770':win_text_set(h,'zhushou')
      if _cls=='Afx:400000:b:10003:6:7fdd13a7':win_text_set(h,'tdx')
      _txt=win_text(h)
      tid,pid = tpid(h)
      act = None
      # if _cls=='ConsoleWindowClass': act = win_close(h)
      # if _cls=='TaskManagerWindow': act = win_close(h)
      if _cls=='mininews': act = win_close(h)
      if _cls=='TXGuiFoundation' and _txt=='腾讯网迷你版':
        act = win_close(h)
      #if _cls=='#32770' and _txt=='多帐户登录设置'
      if _txt=='多帐户登录设置' or _txt=='多帐户批量快速登录':
        act = win_close(h)
      rt = (h, hex(h), _txt, _cls, act, IsWindowVisible(h),pid)
      if debug:
        print(rt)
      return rt
    except Exception as ex:
      print(ex)

  w_a = win_find(ph,cls=cls,txt=txt)
  rt_a = [ _chk(v[0]) for v in w_a ]
  return rt_a

bots = lambda ph=[0],ha=None,cls=None,clsa=None,txt=None:bot(
    ha=wins(ph,ha,cls,txt) if not clsa else None,clsa=clsa)

def ensure_win_notmn(h,sleep_else=0,sleep_mx=0.5):
    flags, state, ptMin, ptMax, rect = rt = win_state(h)
    if state==win32con.SW_SHOWMINIMIZED:
        win_change(h,mx=True)
        time.sleep(sleep_mx)
    else:
        time.sleep(sleep_else)
    return rt
def ensure_win_mx(h,sleep_else=0,sleep_mx=1):
    flags, state, ptMin, ptMax, rect = rt = win_state(h)
    if state!=win32con.SW_SHOWMAXIMIZED:
        win_change(h,mx=True)
        time.sleep(sleep_mx)
    else:
        time.sleep(sleep_else)
    return rt

###############################
def wx_send(who,txt,mode='ctrlv',debug=True):
    w=tryx(lambda:win_find(0,cls='ChatWnd',txt=who)[0][0],False)
    if w:
        rect = win_rect(w)
        win_click(w,1,(60,rect[3]-rect[1]-60))
        time.sleep(0.1)
        paste(w,txt,mode='ctrlv')
        time.sleep(0.1)
        PostMessage(w, win32con.WM_KEYDOWN, win32con.VK_RETURN,0)
    else:
        print('not found wx @',who)
    return myself

def qq_send(who,txt):
    w=tryx(lambda:win_find(0,cls='TXGuiFoundation',txt=who)[0][0])
    if w:
        # mode of type_keys, slow
        input(w, txt)
        SendMessage(w, win32con.WM_KEYDOWN, win32con.VK_RETURN,0)
    else:
        print('not found qq @',who)
    return myself

class probe:
    def __init__(self,ev): self._ev=ev
    def __getattr__(self,k): return self._ev(k)

# for chain call
myself = probe(lambda k:eval(k))
#print(myself)

if __name__ == '__main__':
    test()
    import sys
    if not sys.flags.interactive: mon()



# mypyx - more gist

from megadata.mypy import *

def sizeof(v):
    vv = v # todo fn/function
    size_code = None
    size_code_x = None
    getsizeof = sys.getsizeof
    # typeof = type(v)
    if isinstance(v,types.FunctionType): # user function
      size_code = getsizeof(v.__code__)
      size_code_x = getsizeof(dumps_func(v))
    elif callable(v): print('! non-user-func')
    size_vv = sys.getsizeof(vv)
    size_v = sys.getsizeof(v)
    return size_v,size_vv,size_code,size_code_x

sys_version_info = lambda fix_py2_utf8=False:{'version_info':s2o(o2s(sys.version_info))}

#if __name__!='__main__':
#    # make module callable as class
#    import sys
#    sys.modules[__name__] = __class__
#else:
#    print(__name__)

# control call frequency/rate less than freq_count/freq_time
class freqq:
    def __init__(self, init_q=[],freq_count=500, freq_time=60):
        self.freq_count = freq_count
        self.freq_time = freq_time
        self.q = []
        self.qd = [] 
        for v in init_q: self.do_enq(v)
    def do_deq(self):
        while tryx(lambda: self.qd.pop() if now()-self.qd[0]>=self.freq_time else None,False):
            #log1('.');flush1()
            pass
        if len(self.qd)<self.freq_count:
            #rt = tryx(lambda:self.q.pop(),False)
            rt = tryx(self.q.pop,False)
            if rt: self.qd.append(now())
            return rt
    def do_enq(self,param,deq=False):
        self.q.append(param)
        if deq: return self.do_deq()
    def process(self,thefunc):
        while True:
            len_myq_q = len(self.q)
            len_myq_qd = len(self.qd)
            while True:
                data = self.do_deq()
                if data is None: break
                #print(len_myq_q,len_myq_qd,data)

                try_async(lambda:thefunc(data))
            if len_myq_qd<=0 and len_myq_q<=0: break
            sleep(0.1)

class dict1(dict):# dict[]
    def __getitem__(self,k): return self.get(k)

class dict2(dict):# dict[][]
    def __getitem__(self,k):
        v = self.get(k)
        if v is None: self[k] = v = dict1()
        return v

import numpy as np

pctlog2=lambda df,e=0.0001:np.log2(df.where(df>e).pct_change().replace(np.nan,0)+1)
pctloge=lambda df,e=0.0001:np.log(df.where(df>e).pct_change().replace(np.nan,0)+1)

def tiny_email(user, mypass, sender, receiver, Subject, html, smtp_host='smtp.qq.com', smtp_port=465, Cc='', Bcc='',attachment_data=None,attachment_name=None):
    import smtplib
    from email.mime.text import MIMEText
    if attachment_name and attachment_data:
      from email.mime.multipart import MIMEMultipart
      msg=MIMEMultipart('mixed')
      att=MIMEText(attachment_data,'base64','utf-8')
      att['Content-Type']='application/octet-stream'
      att['Content-Disposition']='attachement; filename='+attachment_name
      msg.attach(att)
    else:
      msg=MIMEText(html,'html','utf-8')
    msg['From'] = sender
    if type(receiver) is str:
        receiver_a = [receiver,]
        receiver_s = receiver
    else:#iterable
        receiver_a = receiver
        receiver_s = ';'.join([str(v) for v in receiver])
    msg['To'] = receiver_s
    msg['Subject']= Subject
    msg['Cc'] = Cc
    msg['Bcc'] = Bcc
    receiver_a += [Cc,Bcc]

    server=smtplib.SMTP_SSL(smtp_host, smtp_port) 
    server.login(user, mypass)
    server.sendmail(sender, receiver_a, msg.as_string())
    server.quit()

###################################### dba wrapper (using sqlalchemy)
#'{scheme}://{user}:{password}@{host}:{port}/{db}?charset={charset}#{fragment}'
#'{scheme}://{netloc}@{host}:{port}/{path}?{query}#{fragment}'
#e.g.'mysql://root:123456@127.0.0.1:3306/?charset=utf8mb4'

class dba:
    _ = {'_':now()}
    def __init__(self,*args,**kwargs):
        self.hash = _hash = hash(self)
        dba._[_hash]={}
        try:
            from urllib.parse import quote_plus,urlparse

            password = kwargs.get('password')
            if password: kwargs['password'] = quote_plus(password)

            dbstr = kwargs.get('dbstr')

            dbstr_file = kwargs.get('dbstr_file')
            if dbstr_file: dbstr = read(dbstr_file).strip()

            if not dbstr:
                if len(args)>0:
                    args0 = args[0]
                    if '://' in args0: dbstr = args0
                    else:
                        dbstr_file = args0
                        dbstr = read(dbstr_file).strip()
            if dbstr:
                dbstr = dbstr.format(**kwargs)
                parsed = urlparse(dbstr)
                scheme = parsed.scheme
            else:
                scheme = kwargs.get('scheme')
            self.scheme = scheme
            debug = kwargs.get('debug')
            if debug:
                print('kwargs=',kwargs)
                print('dbstr=',dbstr)
                dba._[_hash]['dbstr']=dbstr
            from sqlalchemy import create_engine
            self.engine = create_engine(dbstr)
        except Exception as ex:
            print('ex=',ex)

    def __del__(self): dba._.pop(self.hash, None)

    def df(self,sql):
        from pandas import read_sql
        return read_sql(sql,self.engine)

    def yielder(self,sql):
        with self.engine.connect() as conn:
            conn = conn.execution_options(stream_results=True)
            for row in conn.execute(sql): yield row


def get_cmdline_args_kwargs():
  #import argparse
  #parser = argparse.ArgumentParser()
  from argparse import ArgumentParser
  parser = ArgumentParser()

  # first round (TODO, simplify logic later...)
  parsed, unknown = parser.parse_known_args() # this is an 'internal' method
  rt_args = []
  for arg in unknown:
      if arg.startswith(("-", "--")):
          parser.add_argument(arg.split('=')[0], type=str)
      else:
          rt_args.append(arg)
  # again to get the parsed
  parsed, unknown = parser.parse_known_args()
  rt_kwargs = parsed.__dict__
  return rt_args, rt_kwargs




from megadata.mypy import *

def on_quit(*a):
  print('on_quit',*a)
  os._exit(0)
hook_quit(on_quit)

def clt_ipc(argv):
  print('clt_ipc',argv)
  argc = len(argv)
  address = build_address(argv[1], argv[2] if argc>2 else None)
  print('address=',address)

  out={}
  #server= lambda v: ipc(address, v, out=out)
  def server(v):
    return ipc(address,v,out=out)

  import pickle
  for line in sys.stdin:

    if line.startswith('lambda'): # TMP TEST RPC
      fc = eval(line)
      #print('debug fc',fc,fc.__code__)
      print(server(dumps_func(fc)))

    # e.g. /"Adm","ping"
    # have fund like /"Adm","ping",*args,**kwargs
    #  or /api("Adm","ping")
    elif line.startswith('/') and not line.startswith('/api'):
        print(server(f'/api({line[1:]})'))

    # super local lines
    # wanjo: play:
    # ;type(ipc(r'\\.\pipe\ipctest','algodata.dao,20220704'))
    # ;server('algodata.dao,20220704')
    # ;server('(api)')('Adm','ping')
    # ;server("""(api('algodata','dfb',20220704))""").shape
    elif line.startswith(';'):
      print( tryx(lambda:eval(line[1:],globals(),{'server':server})) )
      #print(out.keys())

    # e.g.
    # (api("Adm","ping"))
    # /api("Adm","ping")
    # what-so-ever
    else:
      rt = tryx(lambda:server(line))
      print(type(rt),'=>',rt)

if __name__ == '__main__':
  clt_ipc(argv)



from megadata.mypy import tryx, now, parallel, hook_quit, argv, argc, sys, os,wc

def clt_web(argv):
  print('clt_web',argv)
  def on_quit(*a):
    print('quit',a)
    os._exit(0)

  hook_quit(on_quit)

  port = tryx(lambda:int(argv[1]),False) or '80'

  host = tryx(lambda:argv[2],False) or '127.0.0.1'
  protocol = tryx(lambda:int(argv[3]),False) or 'http'

  address = f'{protocol}://{host}:{port}'

  print(address)

  def send_once(v):
    rt = wc(address, v, timeout=20)
    return rt

  def yield_arg():
    for v in range(1,9999):
      yield v

  for line in sys.stdin: print(send_once(line))

if __name__ == '__main__':
  clt_web(argv)



#-*- coding: utf-8 -*-

# py-coding-simplifer by Wanjo 20221010

from __future__ import print_function # for py2 print()
_print=print

from time import time as now, mktime, sleep

load_time = now()

# core of mypy
#exec('def tryx(l,e=print):\n try:return l()\n except Exception as ex:return ex if True==e else e(ex) if e else None')
def tryx(l,e=print):
    try: return l()
    except Exception as ex: return ex if True==e else e(ex) if e else None

import sys
argv = sys.argv
argc = len(argv)

log1 = sys.stdout.write
flush1 = sys.stdout.flush
log2 = sys.stderr.write
flush2 = sys.stderr.flush

def print1(*args):
  for v in args: log1(f'{v} ')
  log1('\n')
  flush1()

import os
touch_dir = lambda fn:os.makedirs(fn,exist_ok=True)
file_exists = os.path.exists
get_mtime = lambda f:os.path.getmtime(f)

def file_rename(fnold,fnnew=None,try_delete_after=False):
    if try_delete_after:
        fnnew = f'{fnnew}.delete.{now()}'
    tryx(lambda:os.rename(fnold,fnnew))
    if try_delete_after:
        tryx(lambda:os.remove(fnnew))

evalx = lambda s,g=globals():eval(s,g)

flag_py2 = sys.version_info.major==2
sys_import = __import__
sys_reload = __builtins__.reload if flag_py2 else sys_import('importlib').reload
refresh = lambda n:sys_reload(sys_import(n))

class probe:
    def __init__(self,ev): self._ev=ev
    def __getattr__(self,k): return tryx(lambda:self._ev(k))

class probex:
    def __init__(self,ev,debug=False):
      self._ev=ev
      self.debug = debug
    def __getattr__(self,k):
      rt = tryx(lambda:self._ev(k),self.debug)
      if rt is None: rt = tryx(lambda:sys_import(k),self.debug)
      return rt

# from megadata.mypy import mypy
mypy = probex(evalx)

if not flag_py2:
    # https://stackoverflow.com/questions/18466079/change-the-connection-pool-size-for-pythons-requests-module-when-in-threading/22253656#22253656
    def patch_connection_pool(**constructor_kwargs):
        from urllib3 import connectionpool, poolmanager
        class MyHTTPConnectionPool(connectionpool.HTTPConnectionPool):
            def __init__(self, *args,**kwargs):
                kwargs.update(constructor_kwargs)
                super(MyHTTPConnectionPool, self).__init__(*args,**kwargs)
        poolmanager.pool_classes_by_scheme['http'] = MyHTTPConnectionPool
        class MyHTTPSConnectionPool(connectionpool.HTTPSConnectionPool):
            def __init__(self, *args,**kwargs):
                kwargs.update(constructor_kwargs)
                super(MyHTTPSConnectionPool, self).__init__(*args,**kwargs)
        poolmanager.pool_classes_by_scheme['https'] = MyHTTPSConnectionPool
    #tryx(lambda:patch_connection_pool(maxsize=16))

def delx(o,k):
    tryx(lambda:o.__delitem__(k),False)
    tryx(lambda:o.__delattr__(k),False)
    return o

import json
class MyJsonEncoder(json.JSONEncoder):
    def default(self, obj): return tryx(lambda:json.JSONEncoder.default(self,obj))
s2o = lambda s:tryx(lambda:json.loads(s),False)
o2s = lambda o,indent=None:tryx(lambda:json.dumps(o, indent=indent, ensure_ascii=False, cls=MyJsonEncoder))

from pickle import dumps as o2b, loads as b2o

def get_urlopen():
  if flag_py2: from urllib2 import urlopen
  else: from urllib.request import urlopen
  return urlopen

# NOTES: one-off simple web call, not for heavy usage!!
wc=lambda u=None, data=None, m='POST', timeout=10:get_urlopen()(url=u,data=data.encode('utf-8') if isinstance(data,str) else o2s(data).encode('utf-8') if data else None,timeout=timeout).read().decode()

# NOTES: one-off ws call, not for heavy usage!!!
# TODO: need to fix for non-blocking...
def wsc(u,data,lines=1):
  s=data.encode('utf-8') if isinstance(data,str) else o2s(data).encode('utf-8') if data else None

  import websocket
  conn = websocket.create_connection(u)

  conn.send(s)
  if lines==1:
    rt = tryx(conn.recv,False)
  else:
    rta = []
    while True:
      rt = tryx(conn.recv,False)
      if rt is None: break
      rta.append(rt)
    rt = '\n'.join(rta)
  tryx(conn.close,False)
  return rt

# NOTES:
## UP {bytes: ..., None: None, str: encode('utf-8'), else:o2s(...).encode('utf-8')}
## DN {bytes: b2o, else:...} 
def ipc(u,data,authkey=None,out=None):
  s=data.encode('utf-8') if isinstance(data,str) else data if isinstance(data,bytes) else o2s(data).encode('utf-8') if data is not None else None
  #print('dbg.ipc=>',type(data),type(s))
  from multiprocessing.connection import Client
  close = False
  if out is None:
    conn = Client(u)
    close = True
  elif 'conn' in out.keys():
    #print('reuse')
    conn = out['conn'] # try reuse
    if conn is None:
      conn = Client(u)
      out['conn'] = conn
  else: # cache into out
    conn = Client(u)
    out['conn'] = conn
  try:
    conn.send(s)
    rt = conn.recv()
  except Exception as ex:
    if out is not None:
      if 'conn' in out.keys():
        del out['conn']
    rt = None
    tryx(conn.close)
    raise ex
  if close: tryx(conn.close)
  #print('dbg.ipc<=',type(rt))
  if isinstance(rt, bytes):
    rt_loads = b2o(rt)
    #print('dbg.ipc',type(rt_loads),len(rt))
    return rt_loads
  return rt

read = lambda f,m='r',encoding='utf-8':open(f,m,encoding=encoding).read()
# for binary: write(f,s,'wb',None)
write = lambda f,s,m='w',encoding='utf-8':open(f,m,encoding=encoding).write(s)
load = lambda f:s2o(read(f))
save = lambda f,o:write(f,o2s(o))

log = lambda fn,o,ln='\n':write(fn, f'{o if type(o) is str else o2s(o)}{ln}', 'a')

import marshal,types
dumps_func = lambda func:marshal.dumps(func.__code__)
loads_func = lambda codes,ctx,name=None:types.FunctionType(marshal.loads(codes),ctx,name=name)
func2file = lambda fc,fn:write(fn,dumps_func(fc),'wb',None)
file2func = lambda fn,ctx=globals(),name=None:types.FunctionType(marshal.loads(read(fn,'rb',None)),ctx,name=name)

class obj(dict):# dictx
    def __init__(self,pa=None):
        for k in pa or {}:self[k]=pa[k]
    def __getitem__(self,key): return self.get(key)
    def __getattr__(self,key): return self[key]
    def __setattr__(self,k,v): self[k]=v

class objx(dict):#dictxx
    def __init__(self,*args,**kwargs):# TODO for v in args.values(): if type(v) is dict, merge k=>v
        for k,v in kwargs.items():self[k]=v
    def __getitem__(self,key): return self.get(key)
    def __getattr__(self,key): return self[key]
    def __setattr__(self,k,v): self[k]=v

def on_quit_default(*a):
  print(*a)
  os._exit(0)

def hook_quit(on_quit=on_quit_default):
    import signal
    signal.signal(signal.SIGINT, on_quit)
    if sys.platform != 'win32': signal.signal(signal.SIGHUP, on_quit)
    signal.signal(signal.SIGTERM, on_quit)

sgn = lambda v:1 if v>0 else -1 if v<0 else 0
lvl = lambda v,d=0.05:round(v/d-sgn(v)*0.5) #level to zero by d
almost = lambda v1,v2,epsilon=0.0001:abs(v1-v2)<epsilon

# tips: for HH:MM:SS, days should be 1970 ;)
def time_maker(days=0,date=None,outfmt=None,infmt='%Y-%m-%d',
        months=0):
    from datetime import datetime,timedelta
    if date is None: _dt = datetime.now()
    else: _dt = datetime.fromtimestamp(int(date))\
        if infmt=='0' or not infmt\
        else datetime.strptime(str(date),infmt)
    if months>0 or months<0:
        from dateutil.relativedelta import relativedelta
        _dt += relativedelta(months=months)
    _dt += timedelta(days=days)
    if outfmt is None: outfmt = infmt
    if outfmt=='0' or not outfmt:
        return int(mktime(_dt.timetuple()))
    return _dt.strftime(outfmt)

#e.g. acct_num = re_match(r'\D*(\d*)',str(acct))
import re
re_match=lambda p,s,a=re.M|re.I:(re.search(p, s, a) or [None,None])[1]
#e.g. re_replace(r'\D+','?',r'test 1234 ok')
re_replace=lambda p,needle,hay,a=re.M|re.I:re.sub(p, needle, hay, a)

# e.g. attachment_data = read('QR.png','rb',None)
# TODO support list of attachment_data/attachment_name later

import multiprocessing
def parallel(func, a, pool_size=None,chunksize=None,mode='default',map_async=False):
  proc_name = multiprocessing.current_process().name
  if mode in ['loky','multiprocessing'] and proc_name!='MainProcess':
    print(f'WARNING: mode-{mode} on {proc_name}, should kick in __main__')
    return []
  #print(f'proc_name={proc_name}')

  if pool_size is None:
    from os import cpu_count
    pool_size =  cpu_count()

  if mode == 'loky': 
    from joblib import Parallel,delayed
    rt = Parallel(n_jobs=pool_size,backend=mode)(delayed(func)(v) for v in a)
    if not map_async: return rt
  if mode == 'multiprocessing':
    Pool = multiprocessing.Pool
  else: # default is thread mode
    from multiprocessing.dummy import Pool
  if map_async:
    return Pool(pool_size).map_async(func, a, chunksize=chunksize)
  else:
    return Pool(pool_size).map(func, a, chunksize=chunksize)

def mygc():
    import gc
    import sys
    # gc-patch for win32
    if sys.platform=='win32':
        # https://stackoverflow.com/questions/31851848/python-program-memory-in-windows
        import ctypes
        from ctypes import wintypes
        kernel32 = ctypes.WinDLL('kernel32', use_last_error=True)

        def errcheck_bool(result, func, args):
            if not result: raise ctypes.WinError(ctypes.get_last_error())
            return args

        kernel32.GetCurrentProcess.restype = wintypes.HANDLE
        kernel32.SetProcessWorkingSetSize.errcheck = errcheck_bool
        kernel32.SetProcessWorkingSetSize.argtypes = (wintypes.HANDLE,
                                                      ctypes.c_size_t,
                                                      ctypes.c_size_t)
        hProcess = kernel32.GetCurrentProcess()
        kernel32.SetProcessWorkingSetSize(hProcess, -1, -1)
    elif sys.platform=='linux':
        from ctypes import cdll, CDLL
        try:
            cdll.LoadLibrary("libc.so.6")
            libc = CDLL("libc.so.6")
            libc.malloc_trim(0)
        except (OSError, AttributeError):
            libc = None
    gc.collect()
    return len(gc.get_objects())

#def md5(s):return mypy.hashlib.md5(bytes(s,encoding='utf8')).hexdigest()
md5=lambda s:mypy.hashlib.md5(bytes(s,encoding='utf8')).hexdigest()

def yielder(func,wrap=tryx,do_yield=True): yield (wrap(func) if wrap else func())
def yielder_loop(func,wrap=tryx,do_yield=True):
  while True:
    rt = yield from yielder(func, wrap, do_yield)
    if do_yield and rt is not None: yield rt

#import threading
#Thread = threading.Thread
Thread = mypy.threading.Thread
#try_async=lambda func:tryx(lambda:threading.Thread(target=func).start())
#try_async=lambda func:Thread(target=func).start()
def try_async(func): Thread(target=func).start()

#import asyncio
#new_event_loop = asyncio.new_event_loop
from asyncio import new_event_loop
def try_asyncio(func):
  new_event_loop().run_in_executor(None,func)

def build_address(arg1,arg2=None,folder='../tmp/'):
  port = tryx(lambda:int(arg1),False)
  if port is None:
    if sys.platform=='win32':
      host = '.' if arg2 is None else arg2
      address = rf'\\{host}\pipe\{arg1}'
    else:
      address = f'{folder}{arg1}.sck'
  else:
    host = '127.0.0.1' if arg2 is None else arg2
    address = (host,port)
  return address

def systemx(cmd,w=None):
  import subprocess
  with subprocess.Popen(cmd, stdout=subprocess.PIPE, stdin=subprocess.PIPE) as p:
      return p.communicate(input=w.encode() if type(w) is str else w)[0] if w else p.stdout.read()

def use(mdlname,clsname=None):
  rt = sys_import(mdlname)
  if clsname: return tryx(lambda:getattr(rt,clsname))
  return rt

